import java.lang.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;

////////////////////////////////////////////////////////////////////////////
// MAV.java 
// implementation of the MAV positioning algorithm
//
// Variables of class MAV:
// -> position      -- coordinates of the object
// -> Vx, Vy        -- components of the speed vector
// -> mass          -- weight of the MAV
// -> air           -- object representing parameters of the medium
// Variables of method MAV.run():
// -> visRange      -- visibility range of the MAV
// -> distance      -- distance to a neighbor
// -> Fx, Fy        -- components of the force vector
////////////////////////////////////////////////////////////////////////////

public class MAV extends Thread {
  Point2D         position;
  double          Vx, Vy;    // velocity components
  double          mass;      // weight of a particle

  boolean         update;
  boolean         airwave;  
  boolean         interrupted;
  double          Ax, Ay;
  Air             air;
  MathFun         mathFun;

  public final int TIMEX;

  public MAV( Air a, Point2D pos, int sleep ) { 
    super( "MAV thread" );
    air = a; 
    position = pos; 
    Vx = 0.0;
    Vy = 0.0;
    mass = 1.0;
    update = false;
    airwave = false;
    Ax = Ay = 0.0;
    air.drawMAV( position );
    TIMEX = sleep;
    mathFun = new MathFun();
  } 

  public void updatePosition() { update = true; }
  public void interrupt()      { interrupted = true; super.interrupt(); }

  // compute the additional force from the airwave (*)
  public void blast( Point2D center ) { 
    Point2D currPosition = getPosition();

    // compute distance from the center
    double dist2 = Math.pow(center.getX()-currPosition.getX(),2) + 
      Math.pow(center.getY()-currPosition.getY(),2);
    double dist = Math.sqrt( dist2 );

    double A = 500 * air.G * mass / dist2;
    // repulsive
    Ax -= (center.getX()-currPosition.getX()) / dist * A;
    Ay -= (center.getY()-currPosition.getY()) / dist * A;
    airwave = true;
  }

  // disrupt trigonometric function computation
  public void radiationHit() { mathFun.corrupt(); }
  // restore trigonometric function computation
  public void useSoftwareEmulation() { mathFun.restore(); }

  public synchronized Point2D getPosition() { 
    return (Point2D)position.clone();
  }

  // a running MAV collects positions of visible MAVs (actually, all, but the
  // ones that are too far are ignored), and computes the geometric mean of
  // gravitational forces
  public void run() {
    Point2D newPos = new Point2D.Double();
    double visRange = 1.5*air.DIST;              // MAV visibility range
    double G = air.G;
    int dirFlag = (Math.random() < 0.5) ? -1 : 1; // direction of corrective
                                                 // displacement
    while( !interrupted ) { 
      if( update ) {
	update = false;
	air.drawMAV( position );
      }

      // apply friction
      Vx *= 1.0-air.FR;
      Vy *= 1.0-air.FR;

      // get coordinates of neighbor MAVs and compute the force
      double Fx = 0.0;
      double Fy = 0.0;
      Vector neighbors = new Vector();
      for( Enumeration mavs = air.getMAVs(); mavs.hasMoreElements(); ) {
	MAV nextMAV = (MAV)mavs.nextElement();
	if( !nextMAV.equals( this ) ) {
	  Point2D nextPos = nextMAV.getPosition();

	  // compute distance
	  double dist2 = Math.pow(nextPos.getX()-position.getX(),2) + 
	    Math.pow(nextPos.getY()-position.getY(),2);
	  double dist = Math.sqrt( dist2 );

	  if( dist <= visRange ) {
	    // here is where the MAV sees it
	    double distance = dist;

	    // execute positioning algorithm
	    double F = G * mass * nextMAV.mass / dist2;
	    if (F > air.Fmax) F = air.Fmax;

	    // project on axes as appropriate
	    
//  	    // angle-based projection for program checking
//  	    double dX = nextPos.getX()-position.getX();
//  	    double dY = nextPos.getY()-position.getY();
//  	    double theta = mathFun.atan(dY/dX);
//  	    if (dX < 0) theta = theta + Math.PI;
	    
//  	    double trigVal;
//  	    if( dist > air.DIST ) {
//  	      // attractive
//  	      trigVal = mathFun.cos( theta );
//  	      Fx += trigVal * F;
//  	      trigVal = mathFun.sin( theta );
//  	      Fy += trigVal * F;
//  	    } else if (air.getRepulse()) {
//  	      // repulsive
//  	      trigVal = mathFun.cos( theta );
//  	      Fx -= trigVal * F;
//  	      trigVal = mathFun.sin( theta );
//  	      Fy -= trigVal * F;
//  	    } 

	    // normal projection
	    if( dist > air.DIST ) {
	      // attractive
	      Fx += (nextPos.getX()-position.getX()) / dist * F;
	      Fy += (nextPos.getY()-position.getY()) / dist * F;
	    } else if (air.getRepulse()) {
	      // repulsive
	      Fx -= (nextPos.getX()-position.getX()) / dist * F;
	      Fy -= (nextPos.getY()-position.getY()) / dist * F;
	    } 

	    // add to the neighbor count
	    // exclude close ones (nearly same position)
 	    if( dist > 0.25*air.DIST ) {
	      // enter the new neighbor into the neighbors set, if it not there
	      boolean found = false;
	      for( Enumeration nbrs = neighbors.elements(); 
		  nbrs.hasMoreElements(); ) {
		// compute the distance between this and the new neighbor
		Point2D currPos = (Point2D)nbrs.nextElement();
		double nd2 = Math.pow(nextPos.getX()-currPos.getX(),2) + 
 		  Math.pow(nextPos.getY()-currPos.getY(),2);
		double nd = Math.sqrt( nd2 );
		if( nd < 0.1*air.DIST ) {
		  found = true;
		  break;
		}
	      } 
	      if( !found ) neighbors.add( nextPos );
	    }
	  }
	}
      }

      // if there are too few neighbors, increase visibility range
//       if( neighbors.size() < 2 ) {
// 	visRange *= 1.5;
// 	G *= 1.5;
// 	System.out.println( "# I'm all aloone!  " + visRange );
//       } else if( visRange > 1.5*air.DIST ) {
// 	visRange /= 1.5;
// 	G /= 1.5;
//       }

      // if there are too few neighbors at the right distance, get nervous
//       if( (neighbors.size() == 1) || (neighbors.size() == 2) ) {
// 	boolean doCorrection = false;
// 	// compute distance to the first neighbor (used by both)
// 	Point2D nPos = (Point2D)neighbors.firstElement();
// 	// compute distance
// 	double dX=1.0, dY=0.0;
// 	double thetaDisp = 0;
// 	double d1X = nPos.getX()-position.getX();
// 	double d1Y = nPos.getY()-position.getY();
// 	double n1d2 = Math.pow(d1X,2) + Math.pow(d1Y,2);
// 	double n1d = Math.sqrt( n1d2 );
// 	if( Math.abs(n1d - air.DIST) < 0.1*air.DIST ) {
// 	  if( neighbors.size() == 1 ) {
// 	    // with one neighbor, use the direction towards the neighbor
// 	    dX = d1X; dY = d1Y;
// 	    thetaDisp = Math.PI/3;
// 	    doCorrection = true;
// 	  }
// 	  if( neighbors.size() == 2 ) {
// 	    // get the second neighbor
// 	    nPos = (Point2D)neighbors.get(1);
// 	    // compute distance to the second neighbor
// 	    double d2X = nPos.getX()-position.getX();
// 	    double d2Y = nPos.getY()-position.getY();
// 	    double n2d2 = Math.pow(d2X,2) + Math.pow(d2Y,2);
// 	    double n2d = Math.sqrt( n2d2 );

// 	    // compute distance between the two points
// 	    double d3X = d2X - d1X;
// 	    double d3Y = d2Y - d1Y;
// 	    double n3d2 = Math.pow(d3X,2) + Math.pow(d3Y,2);
// 	    double n3d = Math.sqrt( n3d2 );

// 	    // if the two neighbors and the current one form an
// 	    // equilateral triangle, position is stable
// 	    if( (Math.abs(n2d - air.DIST) < 0.1*air.DIST) &&
// 	        (Math.abs(n3d - air.DIST) < 0.1*air.DIST) ) {
// 	      dX = (d2X + d1X)/2;
// 	      dY = (d2Y + d1Y)/2;
// 	      thetaDisp = Math.PI/2;
// 	      doCorrection = true;
// 	    }
// 	  }

// 	  // if all conditions are met, destabilize the MAV
	  
// 	  if( doCorrection ) {
// 	    // compute the angle for movement
// 	    double theta = Math.atan(dY/dX);
// 	    if( dX < 0 ) theta = theta + Math.PI;
// 	    // add the correction angle
// 	    theta += dirFlag*thetaDisp;
// 	    // compute additional displacement
// 	    corrX += air.DIST * Math.cos(theta);
// 	    corrY += air.DIST * Math.sin(theta);
// 	    System.out.println( "# correction by (" + corrX + "," + corrY + ")" );
// 	  }
// 	}
//       }

      if (airwave) {
	// a blast occurred
	Fx += Ax;
	Fy += Ay;	
	// airwave decreases with time
	Ax *= 9/10;
	Ay *= 9/10;
      }

      // compute speed
      Vx += Fx/mass;
      Vy += Fy/mass;
      
      // compute the new position and redraw
      newPos.setLocation( position.getX()+Vx, position.getY()+Vy );
      air.redrawMAV( position, newPos );

      synchronized ( this ) { position.setLocation( newPos ); }
      try { sleep( TIMEX*120 ); }
      catch( Exception e ) {}
    }
  }

  // test test!!!!!
  public void correction() { }
}; // class MAV
    
